from asyncio import coroutine
from email.headerregistry import Address
from email.message import EmailMessage
from typing import Union, Iterable

import pytest

from asphalt.mailer.api import Mailer


class DummyMailer(Mailer):
    def __init__(self):
        super().__init__({})
        self.messages = []

    @coroutine
    def deliver(self, messages: Union[EmailMessage, Iterable[EmailMessage]]):
        messages = [messages] if isinstance(messages, EmailMessage) else messages
        self.messages.extend(messages)


class TestMailer:
    @pytest.fixture
    def kwargs(self):
        return {
            'charset': 'iso-8859-1',
            'plain_body': 'Hello åäö',
            'html_body': '<html><body>Hello åäö</body></html>',
            'to': [Address('Test Recipient', 'test', 'domain.country'), 'test2@domain.country'],
            'cc': [Address('Test CC', 'testcc', 'domain.country'), 'testcc2@domain.country'],
            'bcc': [Address('Test BCC', 'testbcc', 'domain.country'), 'testbcc2@domain.country']
        }

    @pytest.fixture
    def mailer(self):
        return DummyMailer()

    @pytest.mark.parametrize('plain_body, html_body', [
        (True, True), (True, False), (False, True)
    ], ids=['both', 'plain', 'html'])
    def test_create_message(self, mailer, kwargs, plain_body, html_body):
        if not plain_body:
            del kwargs['plain_body']
        if not html_body:
            del kwargs['html_body']
        msg = mailer.create_message(subject='test subject', sender='foo@bar.baz', **kwargs)

        assert msg['From'] == 'foo@bar.baz'
        assert msg['Subject'] == 'test subject'
        assert msg['To'] == 'Test Recipient <test@domain.country>, test2@domain.country'
        assert msg['Cc'] == 'Test CC <testcc@domain.country>, testcc2@domain.country'
        assert msg['Bcc'] == 'Test BCC <testbcc@domain.country>, testbcc2@domain.country'

        if plain_body and html_body:
            assert msg['Content-Type'] == 'multipart/alternative'
            plain_part, html_part = msg.iter_parts()
        elif plain_body:
            plain_part, html_part = msg, None
        else:
            plain_part, html_part = None, msg

        if plain_part:
            assert plain_part['Content-Type'] == 'text/plain; charset="iso-8859-1"'
            assert plain_part.get_content() == 'Hello åäö\n'
        if html_part:
            assert html_part['Content-Type'] == 'text/html; charset="iso-8859-1"'
            assert html_part.get_content() == '<html><body>Hello åäö</body></html>\n'

    def test_add_attachment(self, mailer):
        msg = mailer.create_message(subject='foo')
        mailer.add_attachment(msg, b'binary content', filename='test.dat')
        attachments = list(msg.iter_attachments())
        assert len(attachments) == 1
        assert attachments[0]['Content-Type'] == 'application/octet-stream'
        assert attachments[0]['Content-Disposition'] == 'attachment; filename="test.dat"'

    @pytest.mark.asyncio
    def test_add_file_attachment(self, mailer):
        msg = mailer.create_message(subject='foo')
        yield from mailer.add_file_attachment(msg, __file__)
        attachments = list(msg.iter_attachments())
        assert len(attachments) == 1
        assert attachments[0]['Content-Type'] == 'text/x-python'
        assert attachments[0]['Content-Disposition'] == 'attachment; filename="test_api.py"'

    @pytest.mark.asyncio
    def test_create_and_deliver(self, mailer, kwargs):
        yield from mailer.create_and_deliver(subject='test subject', sender='foo@bar.baz',
                                             **kwargs)
        assert len(mailer.messages) == 1
        assert isinstance(mailer.messages[0], EmailMessage)
        assert mailer.messages[0]['From'] == 'foo@bar.baz'



from asyncio import coroutine
from asyncio.streams import StreamWriter, StreamReader
from asyncio.tasks import wait_for
from typing import Iterable, Union, Dict, Any
from base64 import b64encode
from email.message import EmailMessage
from email.policy import SMTP
import platform
import re

from typeguard import check_argument_types
from asphalt.core.connectors import Connector, create_connector
from asphalt.core.context import Context
from asphalt.core.concurrency import asynchronous

from ..api import Mailer, DeliveryError
from ..util import get_recipients

__all__ = 'SMTPError', 'SMTPMailer'


class SMTPError(DeliveryError):
    """
    Raised when the SMTP server returns a response with a code of 400 or greater.

    :param code: the numeric response code
    :param error: the error message from the server
    """

    def __init__(self, code: int, error: str):
        super().__init__('server returned error: {} {}'.format(code, error))
        self.code = code
        self.error = error


class SMTPConnection:
    __slots__ = 'reader', 'writer', 'timeout'

    response_re = re.compile(r'(\d{3})(-| )(.*)')

    def __init__(self, reader: StreamReader, writer: StreamWriter, timeout: int):
        self.reader = reader
        self.writer = writer
        self.timeout = timeout

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.writer.close()

    @classmethod
    @coroutine
    def create(cls, connector: Connector, timeout: int):
        reader, writer = yield from connector.connect()
        return cls(reader, writer, timeout)

    @coroutine
    def command(self, command: str, *args: str, expect_response: bool=True):
        # Send the command
        data = ' '.join((command,) + args).encode('ascii') + b'\r\n'
        self.write(data)
        if expect_response:
            return (yield from self.get_response())

    @coroutine
    def get_response(self):
        lines = []
        while True:
            read_task = self.reader.readline()
            line = yield from wait_for(read_task, self.timeout)
            if not line.endswith(b'\r\n'):
                raise DeliveryError('server closed connection')

            line = line.decode(errors='replace').rstrip()
            match = self.response_re.match(line)
            if not match:
                raise DeliveryError('SMTP protocol error: received unexpected response: {}'.
                                    format(line))

            code, separator, message = match.groups()
            lines.append(message)
            if separator == ' ':
                code = int(code)
                response = '\n'.join(lines)
                if code >= 400:
                    raise SMTPError(code, response)
                return response

    def write(self, data: bytes):
        self.writer.write(data)


class SMTPMailer(Mailer):
    """
    A mailer that uses the ESMTP protocol (:rfc:`2821`) to send mails.

    Supports only unencrypted or implicit TLS connections.
    STARTTLS is not supported yet.
    Connects to port 25 on ``127.0.0.1`` by default.

    If you don't know what SMTP server address to use, your Internet
    service provider should have provided you their SMTP server address
    when you signed up.

    :param connector: a connector instance or endpoint for
                      :func:`~asphalt.core.connectors.create_connector`
    :param username: user name to authenticate as
    :param password: password to authenticate with
    :param timeout: timeout (in seconds) for connect and read operations
    :param defaults: default values for omitted keyword arguments of
                     :meth:`~asphalt.mailer.api.Mailer.create_message`
    """

    __slots__ = 'connector', 'username', 'password', 'timeout'

    def __init__(self, *, connector: Union[str, Connector]='tcp://127.0.0.1', username: str=None,
                 password: str=None, timeout: Union[int, float]=10, defaults: Dict[str, Any]=None):
        assert check_argument_types()
        super().__init__(defaults)
        self.connector = connector
        self.username = username
        self.password = password
        self.timeout = timeout

    @coroutine
    def start(self, ctx: Context):
        if not isinstance(self.connector, Connector):
            defaults = {'port': 25, 'timeout': self.timeout}
            self.connector = yield from create_connector(self.connector, defaults, ctx)

    @asynchronous
    def deliver(self, messages: Union[EmailMessage, Iterable[EmailMessage]]):
        assert check_argument_types()
        if isinstance(messages, EmailMessage):
            messages = [messages]

        with (yield from SMTPConnection.create(self.connector, self.timeout)) as connection:
            yield from connection.get_response()
            response = yield from connection.command('EHLO', platform.node())
            features = response.split('\n')[1:]
            policy = SMTP.clone(cte_type='8bit' if '8BITMIME' in features else '7bit')
            pipelining = 'PIPELINING' in features

            try:
                # Authenticate if necessary
                if self.username and self.password:
                    auth_mechanisms = []
                    for feature in features:
                        if feature.startswith('AUTH '):
                            auth_mechanisms = feature[5:].split()
                            break

                    if 'PLAIN' in auth_mechanisms:
                        token = b64encode('\0{}\0{}'.format(self.username, self.password).encode())
                        yield from connection.command('AUTH PLAIN', token.decode())
                    elif 'LOGIN' in auth_mechanisms:
                        yield from connection.command('AUTH LOGIN', self.username, self.password)
                    else:
                        raise DeliveryError('server does not support any of our authentication '
                                            'methods')

                # Send each message in a separate mail transaction
                for message in messages:
                    try:
                        recipients = get_recipients(message)
                        yield from connection.command(
                            'MAIL FROM: <{}>'.format(message['From'].addresses[0].addr_spec),
                            expect_response=not pipelining)
                        for recipient in recipients:
                            yield from connection.command('RCPT TO: <{}>'.format(recipient),
                                                          expect_response=not pipelining)

                        yield from connection.command('DATA', expect_response=not pipelining)

                        # If pipelining was in use, the responses to the MAIL, RCPT and DATA
                        # commands will be sent by the server now
                        if pipelining:
                            for _ in range(2 + len(recipients)):
                                yield from connection.get_response()

                        envelope = message.as_bytes(policy=policy)
                        connection.write(envelope + b'\r\n')
                        yield from connection.command('.')
                    except DeliveryError as e:
                        e.message = message
                        raise
            except DeliveryError:
                yield from connection.command('QUIT', expect_response=False)
                raise
            else:
                yield from connection.command('QUIT', expect_response=False)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, str(self.connector))



from asyncio import coroutine
from typing import Dict, Any
import logging

from typeguard import check_argument_types
from asphalt.core.component import Component
from asphalt.core.context import Context
from asphalt.core.util import PluginContainer

from .api import Mailer

mailer_backends = PluginContainer('asphalt.mailer.mailers', Mailer)

logger = logging.getLogger(__name__)


class MailerComponent(Component):
    """
    Provides a way for sending email.

    Publishes one or more :class:`Mailer` compatible objects as resources and context variables.

    If more than one mailer is to be configured, provide a ``mailers`` argument as a dictionary
    where the key is the resource name and the value is a dictionary of keyword arguments to
    :meth:`create_mailer`. Otherwise, directly pass those keyword arguments to the component
    constructor itself.

    :param mailers: a dictionary of Mailer resource name -> :meth:`create_mailer` keyword arguments
    :param default_mailer_args: keyword arguments given to the backend class
    """

    def __init__(self, mailers: Dict[str, Dict[str, Any]]=None, **default_mailer_args):
        assert check_argument_types()
        if mailers and default_mailer_args:
            raise ValueError('specify either a "mailers" dictionary or the default mailer\'s '
                             'options directly, but not both')

        mailers = mailers or {}
        if default_mailer_args:
            default_mailer_args.setdefault('context_attr', 'mailer')
            mailers['default'] = default_mailer_args

        self.mailers = [self.create_mailer(alias, **kwargs) for alias, kwargs in mailers.items()]

    @classmethod
    def create_mailer(cls, resource_name: str, backend: str, context_attr: str=None,
                      **backend_kwargs):
        """
        Configure a Mailer backend with the given parameters.

        :param resource_name: resource name the mailer will be
            published as
        :param backend: specifies the type of mailer
            (declared as entry points in the ``asphalt.mailer.mailers``
            namespace)
        :param context_attr: the mailer's attribute name on the context
            (defaults to the value of ``resource_name``)
        :param backend_kwargs: keyword arguments passed to the
            constructor of the backend class

        """
        assert check_argument_types()
        context_attr = context_attr or resource_name
        mailer = mailer_backends.create_object(backend, **backend_kwargs)
        return resource_name, context_attr, mailer

    @coroutine
    def start(self, ctx: Context):
        for resource_name, context_attr, mailer in self.mailers:
            yield from mailer.start(ctx)
            yield from ctx.publish_resource(mailer, resource_name, context_attr, types=[Mailer])
            logger.info('Configured mailer (%s / ctx.%s)', resource_name, context_attr)



from typing import Iterable, Union, Dict, Any
from abc import abstractmethod, ABCMeta
from asyncio import coroutine
from email.headerregistry import Address
from email.message import EmailMessage
from mimetypes import guess_type
from pathlib import Path

from typeguard import check_argument_types
from asphalt.core.context import Context
from asphalt.core.concurrency import blocking

__all__ = 'DeliveryError', 'Mailer'

AddressListType = Union[str, Address, Iterable[Union[str, Address]]]


class DeliveryError(Exception):
    """
    Raised when there's an error with mail delivery.

    :ivar error: the error message
    :ivar message: the email message related to the failure, if any
    """

    def __init__(self, error: str, message: EmailMessage=None):
        super().__init__(error, message)

    def __str__(self):
        return 'error sending mail message: {}'.format(self.args[0])


class Mailer(metaclass=ABCMeta):
    """
    This is the abstract base class for all mailers.

    :param defaults: default values for omitted keyword arguments of :meth:`create_message`
    """

    __slots__ = 'defaults'

    def __init__(self, defaults: Dict[str, Any]=None):
        self.defaults = defaults or {}

    @coroutine
    def start(self, ctx: Context):
        """
        Perform any necessary setup procedures.

        This is a coroutine. This method is called by the component on initialization.

        :param ctx: the component's context
        """

    def create_message(self, *, subject: str=None, sender: Union[str, Address]=None,
                       to: AddressListType=None, cc: AddressListType=None,
                       bcc: AddressListType=None, charset: str='utf-8', plain_body: str=None,
                       html_body: str=None) -> EmailMessage:
        """
        A convenience method for creating an :class:`~email.message.EmailMessage` to be sent later
        using :meth:`deliver`.

        :param subject: subject line for the message
        :param sender: sender address displayed in the message (the From: header)
        :param to: primary recipient(s) (displayed in the message)
        :param cc: secondary recipient(s) (displayed in the message)
        :param bcc: secondary recipient(s) (**not** displayed in the message)
        :param charset: character encoding of the message
        :param plain_body: plaintext body
        :param html_body: HTML body

        """
        assert check_argument_types()
        msg = EmailMessage()
        msg['Subject'] = subject or self.defaults.get('subject')
        if sender:
            msg['From'] = sender or self.defaults.get('sender')
        if to:
            msg['To'] = to or self.defaults.get('to')
        if cc:
            msg['Cc'] = cc or self.defaults.get('cc')
        if bcc:
            msg['Bcc'] = bcc or self.defaults.get('bcc')

        if plain_body is not None and html_body is not None:
            msg.set_content(plain_body, charset=charset)
            msg.add_alternative(html_body, charset=charset, subtype='html')
        elif plain_body is not None:
            msg.set_content(plain_body, charset=charset)
        elif html_body is not None:
            msg.set_content(html_body, charset=charset, subtype='html')

        return msg

    @classmethod
    def add_attachment(cls, msg: EmailMessage, content: bytes, filename: str,
                       mimetype: str='application/octet-stream'):
        """
        Adds binary data as an attachment to an :class:`~email.message.EmailMessage`.

        :param msg: the message
        :param content: the contents of the attachment
        :param mimetype: the MIME type indicating the type of the file
        :param filename: the displayed file name in the message

        """
        assert check_argument_types()
        maintype, subtype = mimetype.split('/', 1)
        msg.add_attachment(content, maintype=maintype, subtype=subtype, filename=filename)

    @classmethod
    @blocking
    def add_file_attachment(cls, msg: EmailMessage, path: Union[str, Path], filename: str=None,
                            mimetype: str=None):
        """
        Create an attachment on the given :class:`~email.message.EmailMessage` from a file on the
        filesystem.

        This is a coroutine.

        The default value for the ``filename`` argument is the file name part of the given path.
        The default value for the ``mimetype`` argument is guessed from the file name.

        :param msg: the message
        :param path: path to the file to attach
        :param mimetype: the MIME type indicating the type of the file
        :param filename: the displayed file name in the message

        """
        assert check_argument_types()
        path = Path(path)
        with path.open('rb') as f:
            content = f.read()
        if not mimetype:
            mimetype = guess_type(str(path))[0] or 'application/octet-stream'
        if filename is None:
            filename = path.name

        cls.add_attachment(msg, content, filename, mimetype)

    def create_and_deliver(self, **kwargs):
        """
        Build a new email message and deliver it.

        This is a coroutine.

        This is a shortcut to calling :meth:`create_message` and then passing the result to
        :meth:`deliver`.

        :param kwargs: keyword arguments passed to :meth:`create_message`
        """

        msg = self.create_message(**kwargs)
        return self.deliver(msg)

    @abstractmethod
    def deliver(self, messages: Union[EmailMessage, Iterable[EmailMessage]]):
        """
        Deliver the given message(s).

        This is a coroutine.

        :param messages: the message or iterable of messages to deliver
        """



"""
A simple command line tool that connects to an SMTP server, sends a
mail message and then exits.

You can experiment by modifying the arguments to add_component()
and/or ctx.mailer.create_and_deliver().
"""

import asyncio
import logging

from asphalt.core.component import ContainerComponent
from asphalt.core.context import Context
from asphalt.core.runner import run_application


class MailSenderComponent(ContainerComponent):
    @asyncio.coroutine
    def start(self, ctx: Context):
        self.add_component('mailer', backend='smtp', connector='tcp+ssl://smtp.example.org:465',
                           username='myusername', password='secret')
        yield from super().start(ctx)

        yield from ctx.mailer.create_and_deliver(
            subject='Test email', sender='Sender <sender@example.org>',
            to='Recipient <person@example.org>', plain_body='Hello, world!')
        asyncio.get_event_loop().stop()

run_application(MailSenderComponent(), logging=logging.DEBUG)



"""
A simple command line tool that connects to an SMTP server, sends a
mail message with an attachment and then exits.

In order to make it work, you need to fill in the SMTP backend
arguments (connector, username, password) and the .

The program requires one command line argument, the path to the
attachment file.
"""

import asyncio
import logging
import os.path
import sys

from asphalt.core.component import ContainerComponent
from asphalt.core.context import Context
from asphalt.core.runner import run_application


class MailSenderComponent(ContainerComponent):
    def __init__(self, attachment):
        super().__init__()
        self.attachment = attachment

    @asyncio.coroutine
    def start(self, ctx: Context):
        self.add_component('mailer', backend='smtp', connector='tcp+ssl://smtp.example.org:465',
                           username='myusername', password='secret')
        yield from super().start(ctx)

        message = ctx.mailer.create_message(
            subject='Test email with attachment', sender='Sender <sender@example.org>',
            to='Recipient <person@example.org>', plain_body='Take a look at this file!')
        yield from ctx.mailer.add_file_attachment(message, self.attachment)
        yield from ctx.mailer.deliver(message)
        asyncio.get_event_loop().stop()


if len(sys.argv) < 2:
    print('Specify the path to the attachment as the argument to this script!', file=sys.stderr)
    sys.exit(1)

if not os.path.isfile(sys.argv[1]):
    print('The attachment ({}) does not exist or is not a regular file', file=sys.stderr)
    sys.exit(2)

run_application(MailSenderComponent(sys.argv[1]), logging=logging.DEBUG)
